@@ -1,16 +1,19 @@
 #!/usr/bin/env python
 # -*- coding: utf-8 -*-
 
-from json import JSONDecodeError
+import enum
 import logging
 import re
+from copy import deepcopy
 from datetime import datetime, timedelta
-from typing import Any, Dict, List, Optional, Set
+from json import JSONDecodeError
+from typing import Any, Callable, Dict, List, NamedTuple, Optional, Set
 from urllib.error import HTTPError, URLError
 from urllib.parse import quote_plus
 from urllib.request import urlopen
 
 import requests
+from bs4 import BeautifulSoup
 from cdp_backend.database.constants import (
     EventMinutesItemDecision,
     MatterStatusDecision,
@@ -23,14 +26,19 @@ from cdp_backend.pipeline.ingestion_models import (
     Matter,
     MinutesItem,
     Person,
+    Role,
     Session,
     SupportingFile,
     Vote,
-    Role,
 )
 
-from .types import ContentURIs
-from .scraper_utils import IngestionModelScraper, reduced_list, str_simplified
+from .scraper_utils import (
+    IngestionModelScraper,
+    reduced_list,
+    sanitize_roles,
+    str_simplified,
+)
+from .types import ContentURIs, ScraperStaticData
 
 ###############################################################################
 
@@ -105,6 +113,10 @@ LEGISTAR_DATETIME_FORMAT = "%Y-%m-%dT%H:%M:%S"
 
 known_legistar_persons: Dict[int, Dict[str, Any]] = {}
 known_legistar_bodies: Dict[int, Dict[str, Any]] = {}
+# video web page parser type per municipality
+video_page_parser: Dict[
+    str, Callable[[BeautifulSoup], Optional[List[ContentURIs]]]
+] = {}
 
 
 def get_legistar_body(
@@ -369,6 +381,189 @@ def get_legistar_events_for_timespan(
     return response
 
 
+class ContentUriScrapeResult(NamedTuple):
+    class Status(enum.IntEnum):
+        # Web page(s) are in unrecognized structure
+        UnrecognizedPatternError = -1
+        # Error in accessing some resource
+        ResourceAccessError = -2
+        # Video was not provided for the event
+        ContentNotProvidedError = -3
+        # Found URIs to video and optional caption
+        Ok = 0
+
+    status: Status
+    uris: Optional[List[ContentURIs]] = None
+
+
+def get_legistar_content_uris(client: str, legistar_ev: Dict) -> ContentUriScrapeResult:
+    """
+    Return URLs for videos and captions from a Legistar/Granicus-hosted video web page
+
+    Parameters
+    ----------
+    client: str
+        Which legistar client to target. Ex: "seattle"
+    legistar_ev: Dict
+        Data for one Legistar Event.
+
+    Returns
+    -------
+    ContentUriScrapeResult
+        status: ContentUriScrapeResult.Status
+            Status code describing the scraping process. Use uris only if status is Ok
+        uris: Optional[List[ContentURIs]]
+            URIs for video and optional caption
+
+    Raises
+    ------
+    NotImplementedError
+        Means the content structure of the web page hosting session video has changed.
+        We need explicit review and update the scraping code.
+
+    See Also
+    --------
+    LegistarScraper.get_content_uris()
+    """
+    global video_page_parser
+
+    # prefer video file path in legistar Event.EventVideoPath
+    if legistar_ev[LEGISTAR_SESSION_VIDEO_URI]:
+        return (
+            ContentUriScrapeResult.Status.Ok,
+            [
+                ContentURIs(
+                    video_uri=str_simplified(legistar_ev[LEGISTAR_SESSION_VIDEO_URI]),
+                    caption_uri=None,
+                )
+            ],
+        )
+    if not legistar_ev[LEGISTAR_EV_SITE_URL]:
+        return (ContentUriScrapeResult.Status.UnrecognizedPatternError, None)
+
+    try:
+        # a td tag with a certain id pattern.
+        # this is usually something like
+        # https://somewhere.legistar.com/MeetingDetail.aspx...
+        # that is a summary-like page for a meeting
+        with urlopen(legistar_ev[LEGISTAR_EV_SITE_URL]) as resp:
+            soup = BeautifulSoup(resp.read(), "html.parser")
+
+    except (URLError, HTTPError) as e:
+        log.debug(f"{legistar_ev[LEGISTAR_EV_SITE_URL]}: {str(e)}")
+        return (ContentUriScrapeResult.Status.ResourceAccessError, None)
+
+    # this gets us the url for the web PAGE containing the video
+    # video link is provided in the window.open()command inside onclick event
+    # <a id="ctl00_ContentPlaceHolder1_hypVideo"
+    # data-event-id="75f1e143-6756-496f-911b-d3abe61d64a5"
+    # data-running-text="In&amp;nbsp;progress" class="videolink"
+    # onclick="window.open('Video.aspx?
+    # Mode=Granicus&amp;ID1=8844&amp;G=D64&amp;Mode2=Video','video');
+    # return false;"
+    # href="#" style="color:Blue;font-family:Tahoma;font-size:10pt;">Video</a>
+    extract_url = soup.find(
+        "a",
+        id=re.compile(r"ct\S*_ContentPlaceHolder\S*_hypVideo"),
+        class_="videolink",
+    )
+    if extract_url is None:
+        return (ContentUriScrapeResult.Status.UnrecognizedPatternError, None)
+    # the <a> tag will not have this attribute if there is no video
+    if "onclick" not in extract_url.attrs:
+        return (ContentUriScrapeResult.Status.ContentNotProvidedError, None)
+
+    # NOTE: after this point, failing to scrape video url should raise an exception.
+    # we need to be alerted that we probabaly have a new web page structure.
+
+    extract_url = extract_url["onclick"]
+    start = extract_url.find("'") + len("'")
+    end = extract_url.find("',")
+    video_page_url = f"https://{client}.legistar.com/{extract_url[start:end]}"
+
+    log.debug(f"{legistar_ev[LEGISTAR_EV_SITE_URL]} -> {video_page_url}")
+
+    def _parse_format_1(soup: BeautifulSoup) -> Optional[List[ContentURIs]]:
+        # source link for the video is embedded in the script of downloadLinks.
+        # <script type="text/javascript">
+        # var meta_id = '',
+        # currentClipIndex = 0,
+        # clipList = eval([8844]),
+        # downloadLinks = eval([["\/\/69.5.90.100:443\/MediaVault\/Download.aspx?
+        # server=king.granicus.com&clip_id=8844",
+        # "http:\/\/archive-media.granicus.com:443\/OnDemand\/king\/king_e560cf63-5570-416e-a47d-0e1e13652224.mp4",null]]);
+        # </script>
+
+        video_script_text = soup.find("script", text=re.compile(r"downloadLinks"))
+        if video_script_text is None:
+            return None
+
+        video_script_text = video_script_text.string
+        # Below two lines of code tries to extract video url from downLoadLinks variable
+        # "http:\/\/archive-media.granicus.com:443\/OnDemand\/king\/king_e560cf63-5570-416e-a47d-0e1e13652224.mp4"
+        downloadLinks = video_script_text.split("[[")[1]
+        video_url = downloadLinks.split('",')[1].strip('"')
+        # Cleans up the video url to remove backward slash(\)
+        video_uri = video_url.replace("\\", "")
+        # caption URIs are not found for kingcounty events.
+        return [ContentURIs(video_uri=video_uri, caption_uri=None)]
+
+    def _parse_format_2(soup: BeautifulSoup) -> Optional[List[ContentURIs]]:
+        # <div id="download-options">
+        # <a href="...mp4">
+        video_url = soup.find("div", id="download-options")
+        if video_url is None:
+            return None
+        return [ContentURIs(str_simplified(video_url.a["href"]))]
+
+    def _parse_format_3(soup: BeautifulSoup) -> Optional[List[ContentURIs]]:
+        # <video>
+        # <source src="...">
+        # <track src="...">
+        video_url = soup.find("video")
+        if video_url is None:
+            return None
+        return [
+            ContentURIs(
+                video_uri=f"https:{str_simplified(video_url.source['src'])}",
+                caption_uri=(
+                    (
+                        f"http://{client}.granicus.com/"
+                        f"{str_simplified(video_url.track['src'])}"
+                    )
+                    # transcript is nice to have but not required
+                    if video_url.find("track") is not None
+                    and "src" in video_url.track.attrs
+                    else None
+                ),
+            )
+        ]
+
+    with urlopen(video_page_url) as resp:
+        # now load the page to get the actual video url
+        soup = BeautifulSoup(resp.read(), "html.parser")
+
+        if client in video_page_parser:
+            # we alrady know which format parser to call
+            uris = video_page_parser[client](soup)
+        else:
+            for parser in [_parse_format_1, _parse_format_2, _parse_format_3]:
+                uris = parser(soup)
+                if uris is not None:
+                    # remember so we just call this from here on
+                    video_page_parser[client] = parser
+                    break
+            else:
+                uris = None
+
+    if uris is None:
+        raise NotImplementedError(
+            "get_legistar_content_uris() needs attention. "
+            f"Unrecognized video web page HTML structure: {video_page_url}"
+        )
+    return (ContentUriScrapeResult.Status.Ok, uris)
+
+
 class LegistarScraper(IngestionModelScraper):
     """
     Base class for transforming Legistar API data to CDP IngestionModel.
@@ -433,10 +628,8 @@ class LegistarScraper(IngestionModelScraper):
         Regex pattern used to convert Legistar instance's minutes item failure to CDP
         constant value.
         Default: "not|fail"
-    known_persons: Optional[Dict[str, Person]]
-        Dictionary used to inject information into Persons in place of missing
-        attributes after dynamic scraping in get_events().
-        Default: None
+    static_data: Optional[ScraperStaticData]
+        predefined Seats, Bodies and Persons used to provide more accurate Person.seat.
     person_aliases: Optional[Dict[str, Set[str]]]
         Dictionary used to catch name aliases
         and resolve improperly unique Persons to the one correct Person.
@@ -470,7 +663,7 @@ class LegistarScraper(IngestionModelScraper):
         matter_rejected_pattern: str = r"rejected|dropped",
         minutes_item_decision_passed_pattern: str = r"pass",
         minutes_item_decision_failed_pattern: str = r"not|fail",
-        known_persons: Optional[Dict[str, Person]] = None,
+        static_data: Optional[ScraperStaticData] = None,
         person_aliases: Optional[Dict[str, Set[str]]] = None,
         role_replacements: Optional[Dict[str, str]] = None,
     ):
@@ -499,7 +692,7 @@ class LegistarScraper(IngestionModelScraper):
             minutes_item_decision_failed_pattern
         )
 
-        self.known_persons = known_persons
+        self.static_data = static_data
         self.role_replacements = role_replacements or {}
 
     def get_matter_status(self, legistar_matter_status: str) -> Optional[str]:
@@ -773,7 +966,7 @@ class LegistarScraper(IngestionModelScraper):
             From Legistar OfficeRecords. None if missing information.
         """
         if not legistar_office_records:
-            return None
+            legistar_office_records = []
 
         return reduced_list(
             [
@@ -1221,51 +1414,61 @@ class LegistarScraper(IngestionModelScraper):
         --------
         cdp_scrapers.legistar_utils.get_legistar_events_for_timespan
         """
-        log.critical(
-            "get_content_uris() is required because "
-            f"Legistar Event.EventVideoPath is not used by {self.client_name}"
+        # see if our base legistar/granicus video parsing routine will work
+        result, uris = get_legistar_content_uris(self.client_name, legistar_ev)
+        if result in [
+            ContentUriScrapeResult.Status.Ok,
+            ContentUriScrapeResult.Status.ContentNotProvidedError,
+        ]:
+            return uris or []
+
+        raise NotImplementedError(
+            f"Please provide get_content_uris() for {self.client_name}"
         )
-        raise NotImplementedError
 
     def inject_known_person(self, person: Person) -> Person:
         """
-        Inject information from self.known_persons
-        if person exists in the dictionary for long-term static data
+        Inject information if person exists in static_data.persons
 
         Parameters
         ----------
         person: Person
-            Person into which to inject data from known_persons
+            Person into which to inject data from static_data
 
         Returns
         -------
         Person
-            Input person updated with more information
-            or as-is if person not found in known_persons
+            Input person updated with information from static_data,
+            and seat.roles sanitized.
+
+        See Also
+        --------
+        scraper_utils.sanitize_roles()
         """
         try:
-            known_person = self.known_persons[person.name]
-        except (TypeError, KeyError):
+            known_person = self.static_data.persons[person.name]
+        except (AttributeError, KeyError):
             return person
 
         for attr in person.__dataclass_fields__.keys():
             static_info = getattr(known_person, attr)
             if static_info is not None:
-                # have long-term information provided in "static*.json"; just use it
-                setattr(person, attr, static_info)
-            else:
-                # use dynamically obtained information, which may be None
-                setattr(person, attr, getattr(person, attr))
+                # have long-term information provided in "static*.json"
+                setattr(person, attr, deepcopy(static_info))
 
         # now that we have seat from static hard-coded data
         # we can bring in seat.roles (OfficeRecords from Legistar API)
-        if person.seat and not person.seat.roles:
-            person.seat.roles = self.get_roles(
-                get_legistar_person(
-                    client=self.client_name,
-                    person_id=person.external_source_id,
-                    use_cache=True,
-                )[LEGISTAR_PERSON_ROLES]
+        if person.seat is not None:
+            person.seat.roles = sanitize_roles(
+                person_name=person.name,
+                roles=self.get_roles(
+                    legistar_office_records=get_legistar_person(
+                        client=self.client_name,
+                        person_id=person.external_source_id,
+                        use_cache=True,
+                    )[LEGISTAR_PERSON_ROLES]
+                ),
+                static_data=self.static_data,
             )
 
         return person
@@ -1275,7 +1478,7 @@ class LegistarScraper(IngestionModelScraper):
     ) -> List[EventIngestionModel]:
         """
         Augment with long-term static data that changes very infrequently.
-        e.e. self.known_persons which includes Person.picture_uri, Person.seat
+        e.e. self.static_data which includes Person.picture_uri, Person.seat
 
         Parameters
         ----------
@@ -1287,8 +1490,8 @@ class LegistarScraper(IngestionModelScraper):
         events: List[EventIngestionModel]
             Input events with static information possibly injected
         """
-        # don't waste time if we don't have any static person info at all
-        if not self.known_persons:
+        # don't waste time if we don't have any info at all
+        if not self.static_data:
             return events
 
         for event in events:
@@ -1373,20 +1576,9 @@ class LegistarScraper(IngestionModelScraper):
                     legistar_ev[LEGISTAR_SESSION_TIME],
                 )
             )
-            # prefer video file path in legistar Event.EventVideoPath
-            if legistar_ev[LEGISTAR_SESSION_VIDEO_URI]:
-                list_uri = [
-                    ContentURIs(
-                        video_uri=str_simplified(
-                            legistar_ev[LEGISTAR_SESSION_VIDEO_URI]
-                        ),
-                        caption_uri=None,
-                    )
-                ]
-            else:
-                list_uri = self.get_content_uris(legistar_ev) or [
-                    ContentURIs(video_uri=None, caption_uri=None)
-                ]
+            list_uri = self.get_content_uris(legistar_ev) or [
+                ContentURIs(video_uri=None, caption_uri=None)
+            ]
 
             ingestion_models.append(
                 self.get_none_if_empty(
